import matplotlib.pyplot as plt
from matplotlib import pyplot
from shapely.geometry import Polygon, Point, LineString
from descartes.patch import PolygonPatch
from shapely.ops import linemerge, unary_union, polygonize_full
import parameters as p
import additionalFunctions as a
import json

folder = r"ClassLevelImages/"

# --------------------------- CLASSES
def get_color(minCoef, maxCoef, curItem):
    if maxCoef != minCoef:
        color = 255 * (curItem - minCoef) / (maxCoef - minCoef)
    else:
        color = 255
    c = '#%02X%02X%02X' % (abs(int(255-color)), abs(int(color)),0)
    return c

class MapImage:
    def __init__(self, title):
        fig = pyplot.figure(figsize=(15, 15), dpi=90)
        ax = fig.add_subplot(111)
        ax.set_title(title)
        xrange = [0, 1000]
        yrange = [0, 1000]
        ax.set_xlim(*xrange)
        ax.set_ylim(*yrange)
        ax.set_aspect(1)
        ax.invert_yaxis()
        self.title = title
        self.ax = ax
        self.fig = fig

    def save(self):
        plt.savefig(folder + self.title)
        plt.close()

class Field:
    def __init__(self, polygon, coefficient, pesticides):
        self.polygon = polygon
        self.coefficient = coefficient
        self.pesticides = pesticides

        self.polygonBigCoef = None
        self.polygonSmallCoef = None
        self.polygonBigActive = None
        self.polygonSmallActive = None
        self.allRoadsCoef = 1    # all type of polygons between roads
        self.currentRoadCoef = 0   # coef near roads
        self.nearRoadsCoef=0

    def getHoney(self):
        return self.polygon.area * self.coefficient * p.delta

    def getHoneyWithPesticides(self):
        coef = self.coefficient + self.pesticides
        if coef > 0:
            return self.polygon.area * coef * p.delta
        else:
            return 0

    def getHoneyWithAllRoadsCoef(self):
        coef = self.coefficient * self.allRoadsCoef
        return self.polygon.area * coef * p.delta

    def getHoneyWithAllRoadsCoefAndPesticides(self):
        coef = (self.coefficient + self.pesticides) * self.allRoadsCoef
        if coef < 0:
            coef = 0
        return self.polygon.area * coef * p.delta

    def getHoneyWithAllRoadsCoefAndPesticidesAndNearRoads(self):

        coef = (self.coefficient + self.pesticides) * self.nearRoadsCoef * -self.currentRoadCoef
        if coef < 0:
            coef = 0
        return self.polygon.area * coef * p.delta


    def get_patch(self, facecolor=a.areaRandomColors()):
        patch = PolygonPatch(self.polygon, facecolor=facecolor,
                                     edgecolor=a.v_color(self.polygon), alpha=0.5, zorder=2)
        return patch

    def getArea(self):
        return self.polygon.area

    def save_to_image(self, title):
        img = MapImage(title)
        img.ax.add_patch(self.get_patch())

        img.save()

    def intersection_with_shape(self, shape):
        rez = []
        int_poly = self.polygon.intersection(shape.polygon)
        if int_poly.area != 0:
            int_field = Field(int_poly, self.coefficient, self.pesticides)
            rez.append(int_field)
        int_field_list = FieldsList([], [], [])
        int_field_list.polygonsList = rez
        return int_field_list

class FieldsList():
    def __init__(self, polygonsArray, coefFields, coefPesticides):

        fields_list = []
        for poly, cf, cp in zip(polygonsArray, coefFields, coefPesticides):
            if type(poly) != Polygon:
                pol = Polygon(poly)
            else:
                pol = poly

            current_field = Field(pol, cf, cp)
            fields_list.append(current_field)
        self.polygonsList = fields_list

    def save_to_image(self, title, coef = None):
        img = MapImage(title)

        if coef != None:
            arr = [e.__getattribute__(coef) for e in self.polygonsList]
            minCoef = min(arr)
            maxCoef = max(arr)

        for i, field in enumerate(self.polygonsList):
            if coef != None:
                col = get_color(minCoef, maxCoef, arr[i])
            else:
                col = a.areaRandomColors()
            patch = field.get_patch(facecolor=col)
            img.ax.add_patch(patch)

        img.save()

    def intersection_with_shape(self, shape):
        rez = []
        for field in self.polygonsList:
            int_poly = field.polygon.intersection(shape.polygon)
            if int_poly.area != 0:
                int_field = Field(int_poly, field.coefficient, field.pesticides)
                rez.append(int_field)
        int_field_list = FieldsList([], [], [])
        int_field_list.polygonsList = rez
        return int_field_list

class Road:
    def __init__(self, coordinates, coefficient):
        self.coordinates = coordinates
        self.coefficient = coefficient
        self.currentCoefficient = coefficient
        if self.coefficient < -0.2:
            self.roadType = "Big"
        else:
            self.roadType = "Small"
        line = LineString(coordinates)
        self.polygonWithoutWidth = line.buffer(1, cap_style=1)
        self.polygon = line.buffer(coefficient * (-100), cap_style=1)

    def get_patch(self, facecolor = a.areaRandomColors(), use_width=True):
        if use_width:
            poly = self.polygon
        else:
            poly = self.polygonWithoutWidth
        patch = PolygonPatch(poly, facecolor=facecolor, edgecolor=a.v_color(poly), alpha=0.5, zorder=2)
        return patch

    def intersection_with_shape(self, shape):
        rez = []
        int_poly = self.polygon.intersection(shape.polygon)
        if int_poly.area != 0:
            int_road = Road(self.coordinates, self.coefficient)
            int_road.polygon = int_poly
            rez.append(int_road)

        int_road_list = RoadsList([], [])
        int_road_list.polygonsList = rez
        return int_road_list



class RoadsList:
    def __init__(self, coord_array, coef_array):
        roads_list = [ ]
        for coord, coef in zip(coord_array, coef_array):
            roads_list.append(Road(coord, coef))

        self.polygonsList = roads_list

    def save_to_image(self, title, coef = None):
        img = MapImage(title)
        if coef != None:
            arr = [e.__getattribute__(coef) for e in self.polygonsList]
            if len(arr) > 1:
                minCoef = min(arr)
                maxCoef = max(arr)
            else:
                minCoef = 1
                maxCoef = 2
        for i, road in enumerate(self.polygonsList):
            if coef != None:
                col = get_color(minCoef, maxCoef, arr[i])
                patch = road.get_patch(col, use_width=True)
            else:
                patch = road.get_patch(use_width=False)
            img.ax.add_patch(patch)
        img.save()

    def intersection_with_shape(self, shape):
        rez = []
        for road in self.polygonsList:
            int_poly = road.polygon.intersection(shape.polygon)
            if int_poly.area != 0:
                int_road = Road(road.coordinates, road.coefficient)
                int_road.polygon = int_poly
                rez.append(int_road)
        int_road_list = RoadsList([], [])
        int_road_list.polygonsList = rez
        return int_road_list

class PolygonBetweenRoads:
    def __init__(self, polygon, roadsType):
        self.polygon = polygon
        self.currentCoefficient = None
        self.active = None
        self.roadsType = roadsType

    def get_patch(self, facecolor = a.areaRandomColors()):
        patch = PolygonPatch(self.polygon, facecolor=facecolor, edgecolor="#000000", alpha=0.5, zorder=2)
        return patch

    def intersection_with_shape(self, shape):
        rez = []
        # int_poly_roads= None
        int_poly = self.polygon.intersection(shape.polygon)
        if int_poly.area != 0:
            int_poly_roads = PolygonBetweenRoads(int_poly, self.roadsType)
            int_poly_roads.active = self.active
            int_poly_roads.currentCoefficient = self.currentCoefficient
            rez.append(int_poly_roads)

        int_poly_roads_list = RoadsPolygonList([])
        int_poly_roads_list.polygonsList= rez
        return int_poly_roads_list


    def save_to_image(self, title):
        img = MapImage(title)
        img.ax.add_patch(self.get_patch())
        img.save()

    # ----------------------------------

class RoadsPolygonList():
    def __init__(self, polygonsArray):

        polygon_list = []
        for i in polygonsArray:
            poly = i.polygon
            if type(poly) != Polygon:
                pol = Polygon(poly)
            else:
                pol = poly
            current_road_poly = PolygonBetweenRoads(pol, i.roadsType)
            current_road_poly.active = i.active
            current_road_poly.currentCoefficient = i.currentCoefficient
            current_road_poly.roadsType = i.roadsType

            polygon_list.append(current_road_poly)
        self.polygonsList = polygon_list


    def save_to_image(self, title, coef = None):
        img = MapImage(title)

        for i, field in enumerate(self.polygonsList):
            if coef != None:

                if field.active:
                    col = "red"
                else:
                    col = "silver"
            else:
                col = a.areaRandomColors()
            patch = field.get_patch(facecolor=col)
            img.ax.add_patch(patch)
        img.save()


    def intersection_with_shape(self, shape):
        rez = []
        for field in self.polygonsList:
            int_poly = field.polygon.intersection(shape.polygon)
            if int_poly.area != 0:
                int_field = PolygonBetweenRoads(int_poly, field.roadsType)
                int_field.active = field.active
                int_field.currentCoefficient = field.currentCoefficient
                rez.append(int_field)
        int_field_list = RoadsPolygonList([])
        int_field_list.polygonsList = rez
        return int_field_list

    # --------------------------------------------------------


class PolygonsBetweenRoadsList:
    def __init__(self, allRoads, roadsType):
        road_coords = []
        for road in allRoads.polygonsList:
            if roadsType == "Small" or road.roadType == roadsType:
                road_coords.append(road.coordinates)

        tmp = p.backArray, p.backArrayPlus
        roadsArrayFull = road_coords + list(tmp)
        merged_lines = linemerge(roadsArrayFull)
        border_lines = unary_union(merged_lines)
        decomposition = polygonize_full(border_lines)

        polygons = []
        for dec in decomposition[0]:  # decomposition[0] --- all polygons   /// decomposition[2] --- linestrings
            coord = list(dec.exterior.coords)
            poly = Polygon(coord)
            polygon_between_roads = PolygonBetweenRoads(poly, roadsType)
            polygons.append(polygon_between_roads)

        self.polygonsList = polygons
        self.roadsType = roadsType

    def save_to_image(self, title, show_active=False):
        img = MapImage(title)
        colorTrue = a.areaRandomColors()
        colorFalse = a.areaRandomColors()

        for poly in self.polygonsList:
            if show_active:
                if poly.active:
                    col = colorTrue
                else:
                    col = colorFalse
            else:
                col = a.areaRandomColors()
            img.ax.add_patch(poly.get_patch(col))

        img.save()

    def SetActivePolygonCoef(self, x, y):
        point = Point(x, y)
        if self.roadsType == "Big":
            polygonCoefficient = p.notOnePoligonBetweenBigRoadsCoef
            activeCoefficient = 0
        else:
            polygonCoefficient = p.notOnePoligonBetweenRoadsCoef
            activeCoefficient = 1

        for poly_between_roads in self.polygonsList:
            if poly_between_roads.polygon.contains(point):
                poly_between_roads.active = True
                poly_between_roads.currentCoefficient = activeCoefficient
            else:
                poly_between_roads.active = False
                poly_between_roads.currentCoefficient = polygonCoefficient


class Shape:
    def __init__(self, polygon, x, y):
        self.polygon = polygon
        self.x = x
        self.y = y

    def get_patch(self, facecolor = a.areaRandomColors()):
        patch = PolygonPatch(self.polygon, facecolor=facecolor, edgecolor=a.v_color(self.polygon), alpha=0.5, zorder=2)
        return patch

    def save_to_image(self, title):
        img = MapImage(title)
        img.ax.add_patch(self.get_patch())
        img.ax.plot(self.x, self.y, 'o', color='#000000', zorder=10)
        te = "(" + str(self.x) + "," + str(self.y) + ")"
        img.ax.text(self.x+10, self.y, te, fontsize=14, color='#000000')
        img.save()
#  -------------------------------------------------------------------------------


class Poly:
    def __init__(self, polygon):
        self.polygon = polygon

class PolygonsList():
    def __init__(self, polygonsArray):
        self.polygonsList = createPolygonsFromList(polygonsArray)

    def save_to_image(self, title, facecolor = a.areaRandomColors(), coef = None):
        img = MapImage(title)

        if coef != None:
            arr = [e.__getattribute__(coef) for e in self.polygonsList]
            minCoef = min(arr)
            maxCoef = max(arr)

        for i, field in enumerate(self.polygonsList):
            if coef != None:
                facecolor = get_color(minCoef, maxCoef, arr[i])

            patch = field.get_patch(facecolor = facecolor)
            img.ax.add_patch(patch)

        img.save()


def createPolygonsFromList(polygon):
    List = []
    for i in range(len(polygon)):
        if type(polygon[i]) != Polygon:
            pol = Polygon(polygon[i])
        else:
            pol = polygon[i]
        currentPoly = Poly(pol)
        List.append(currentPoly)
    return List




def intersection_between_fields_and_roads(inter1, inter2):
    rez = []
    for field in inter1.polygonsList:

        for poly in inter2.polygonsList:
            int_poly = field.polygon.intersection(poly.polygon)
            if int_poly.area != 0:

                int_field = Field(int_poly, field.coefficient, field.pesticides)
                int_field.currentRoadCoef = poly.currentCoefficient
                int_field.nearRoadsCoef = field.nearRoadsCoef

                rez.append(int_field)

    int_field_list = FieldsList([], [], [])
    int_field_list.polygonsList = rez
    return int_field_list


def intersection_between_fields_and_road_sum_polygons_intersection_results(inter1, inter2):
    rez = []
    for field in inter1.polygonsList:

        for poly in inter2.polygonsList:
            int_poly = field.polygon.intersection(poly.polygon)
            if int_poly.area != 0:

                int_field = Field(int_poly, field.coefficient, field.pesticides)
                int_field.polygonSmallActive = poly.active
                int_field.allRoadsCoef = poly.currentCoefficient

                int_field.nearRoadsCoef = poly.currentCoefficient

                rez.append(int_field)

    int_field_list = FieldsList([], [], [])
    int_field_list.polygonsList = rez
    return int_field_list


def intersection_between_intersection_polygons(inter1, inter2):
    rez = []
    for poly1 in inter1.polygonsList:

        for poly2 in inter2.polygonsList:

            int_poly = poly1.polygon.intersection(poly2.polygon)
            if int_poly.area != 0:

                if type(int_poly) == Polygon:

                    int_field = PolygonBetweenRoads(int_poly, None)
                    int_field.currentCoefficient = round(poly2.currentCoefficient + poly1.currentCoefficient, 2)
                    int_field.active = poly2.active
                    rez.append(int_field)
                else:

                    if int_poly.geom_type == 'MultiPolygon':

                        for po in int_poly:

                            int_field = PolygonBetweenRoads(po, None)
                            int_field.currentCoefficient = round(poly2.currentCoefficient + poly1.currentCoefficient, 2)
                            int_field.active = poly2.active
                            rez.append(int_field)
                    if int_poly.geom_type == 'GeometryCollection':
                            for po in int_poly:
                                if type(po) == Polygon:

                                    int_field = PolygonBetweenRoads(po, None)
                                    int_field.currentCoefficient = round(
                                        poly2.currentCoefficient + poly1.currentCoefficient, 2)
                                    int_field.active = poly2.active
                                    rez.append(int_field)

    int_field_list = RoadsPolygonList([])
    int_field_list.polygonsList = rez
    return int_field_list

def pointOnRoad(x,y,roads):
    flag = False
    point = Point(y,x)
    for b in range(len(p.roadsArray)):
        line = LineString(p.roadsArray[b])
        if point.within(line):
            print("Point (", x, y, ") is on the Road!!!")
            flag = True
            break
    if flag == False:
        for b in range(len(roads.polygonsList)):
            if point.within(roads.polygonsList[b].polygon):
                print("Point (", x, y, ") is near the Road!!!")
                flag = True
                break
    return flag

# --- JSON records ---
def gen_record(x, y, onRoad, honey, hives):
    data = {
        'x': x,
        'y':y,
        'onRoad': onRoad,
        'honey': honey,
        'hives': hives
    }
    return data

def write_json(data_dict):
    try:
        d = json.load(open("data_file.json"))
    except:
        d = []
    d.append(data_dict)
    with open("data_file.json", "w") as file:
        json.dump(d, file, indent=2)

def write_json_screen(x, y, onRoad, honey, hives):
    d= gen_record(x, y, onRoad, honey, hives)

    print(json.dumps(d))

    print(json.dumps(d, indent=2))
